"""
A video controller for ai2thor
Basic example:
from ai2thor.controller import VideoController
with VideoController() as vc:
    vc.play(vc.MoveAhead())
    vc.wait(5)
    vc.play(vc.MoveAhead())
    vc.exportVideo('thor.mp4')
Known issues:
- Multi agent rotations don't work (since TeleportFull breaks when passing in an AgentID)
"""

from ai2thor.controller import Controller
import cv2
import os
from PIL import Image
import math
from math import erf, sqrt


class VideoController(Controller):
    def __init__(
        self,
        cam_rot=dict(x=85, y=225, z=0),
        cam_pos=dict(x=-1.25, y=7.0, z=-1.0),
        cam_fov=60,
        **controller_kwargs,
    ):
        super().__init__(continuous=True, **controller_kwargs)
        self.step(
            action="AddThirdPartyCamera",
            rotation=initial_camera_rotation,
            position=initial_camera_position,
            fieldOfView=initial_camera_fov,
        )

        self.saved_frames = []
        self.ceiling_off = False
        self.initial_cam_rot = cam_rot.copy()
        self.initial_cam_pos = cam_pos.copy()
        self.initial_cam_fov = cam_fov

    def reset(self, scene):
        """Changes the scene and adds a new third party camera to the initial position."""
        super().reset(scene)
        self.step(
            action="AddThirdPartyCamera",
            rotation=self.initial_cam_rot,
            position=self.initial_cam_pos,
            fieldOfView=self.initial_cam_fov,
        )

    def play(self, *action_generators):
        """Apply multiple actions at the same time (e.g., move multiple agents,
        and pan the camera around the scene.

        Examples
        vc.play(vc.moveAhead())
        vc.wait(60)
        vc.play(vc.moveAhead(), vc.orbitCameraAnimation(0, 0, 0))"""
        # action_generators should be a list of generators (e.g., moveAhead(<Params>))
        # this does many transformations at the same time
        while True:
            # execute next actions if available
            next_actions = [next(generator, False) for generator in action_generators]

            # add the frame to the saved frames after all actions execute
            self.saved_frames.append(self.last_event.third_party_camera_frames[0])

            # remove actions with finished iterators
            next_actions = [action for action in next_actions if action != False]

            if not next_actions:
                # exit after all generators have finished
                break

    def _wait(self, frames=60):
        """Yields a generator used in self.wait()"""
        for _ in range(frames):
            yield self.step(action="Pass")

    def wait(self, frames=60):
        """Do absolutely nothing to the agent. Keep the current frame still, as is.

        Params
        - frames (int)=60: The duration of the do nothing action.
          Note: videos are typically 30fps.

        Example: vc.wait(60)"""
        self.play(self._wait(frames))

    def ToggleCeiling(self):
        """Hides the ceiling. This method is greatly preferred over calling
        step(action='ToggleMapView') directly, since it allows for automatic
        ceiling toggles in the future. (e.g., if the camera is above the
        height of the room, toggle off the ceiling, and vice versa."""
        self.ceiling_off = not self.ceiling_off
        return self.step(action="ToggleMapView")

    def _cdf(self, x, std_dev=0.5, mean=0.0):
        """Cumulative distribution function"""
        return (1.0 + erf((x - mean) / sqrt(2.0 * std_dev ** 2))) / 2.0

    def _linear_to_smooth(self, curr_frame, total_frames, std_dev=0.5, min_val=3):
        # start at -3 STD on a normal gaussian, go to 3 STD on gaussian
        # curr frame should be 1 indexed, and end with total_frames
        assert min_val > 0, "Min val should be > 0"

        if curr_frame == total_frames:
            # removes drifting
            return 1

        return self._cdf(
            -min_val + 2 * min_val * (curr_frame / total_frames), std_dev=std_dev
        )

    def _move(self, actionName, moveMagnitude, frames, smoothAnimation, agentId=None):
        """Yields a generator full of move commands to move the agent incrementally.
        Used as a general move command for MoveAhead, MoveRight, MoveLeft, MoveBack."""
        last_moveMag = 0
        for i in range(frames):
            # smoothAnimation = False => linear animation
            if smoothAnimation:
                next_moveMag = (
                    self._linear_to_smooth(i + 1, frames, std_dev=1) * moveMagnitude
                )
                if agentId is None:
                    yield self.step(
                        action=actionName, moveMagnitude=next_moveMag - last_moveMag
                    )
                else:
                    yield self.step(
                        action=actionName,
                        moveMagnitude=next_moveMag - last_moveMag,
                        agentId=agentId,
                    )
                last_moveMag = next_moveMag
            else:
                if agentId is None:
                    yield self.step(
                        action=actionName, moveMagnitude=moveMagnitude / frames
                    )
                else:
                    yield self.step(
                        action=actionName,
                        moveMagnitude=moveMagnitude / frames,
                        agentId=agentId,
                    )

    def _rotate(self, direction, rotateDegrees, frames, smoothAnimation, agentId=None):
        """Yields a generator full of step(action='TeleportFull') commands to rotate the agent incrementally."""
        if agentId is not None:
            raise ValueError("rotations do not yet work with multiple agents")

        # make it work for left and right rotations
        direction = direction.lower()
        assert direction == "left" or direction == "right"
        if direction == "left":
            rotateDegrees *= -1

        # get the initial rotation
        y0 = self.last_event.metadata["agent"]["rotation"]["y"]
        for i in range(frames):
            # keep the position the same
            p = self.last_event.metadata["agent"]["position"]
            if smoothAnimation:
                if agentId is None:
                    yield self.step(
                        action="TeleportFull",
                        rotation=y0
                        + rotateDegrees
                        * self._linear_to_smooth(i + 1, frames, std_dev=1),
                        agentId=agentId,
                        **p,
                    )
                else:
                    yield self.step(
                        action="TeleportFull",
                        rotation=y0
                        + rotateDegrees
                        * self._linear_to_smooth(i + 1, frames, std_dev=1),
                        **p,
                    )
            else:
                if agentId is None:
                    yield self.step(
                        action="TeleportFull",
                        rotation=y0 + rotateDegrees * ((i + 1) / frames),
                        agentId=agentId,
                        **p,
                    )
                else:
                    yield self.step(
                        action="TeleportFull",
                        rotation=y0 + rotateDegrees * ((i + 1) / frames),
                        **p,
                    )

    def MoveAhead(self, moveMagnitude=1, frames=60, smoothAnimation=True, agentId=None):
        return self._move(
            "MoveAhead", moveMagnitude, frames, smoothAnimation, agentId=agentId
        )

    def MoveBack(self, moveMagnitude=1, frames=60, smoothAnimation=True, agentId=None):
        return self._move(
            "MoveBack", moveMagnitude, frames, smoothAnimation, agentId=agentId
        )

    def MoveLeft(self, moveMagnitude=1, frames=60, smoothAnimation=True, agentId=None):
        return self._move(
            "MoveLeft", moveMagnitude, frames, smoothAnimation, agentId=agentId
        )

    def MoveRight(self, moveMagnitude=1, frames=60, smoothAnimation=True, agentId=None):
        return self._move(
            "MoveRight", moveMagnitude, frames, smoothAnimation, agentId=agentId
        )

    def RotateRight(
        self, rotateDegrees=90, frames=60, smoothAnimation=True, agentId=None
    ):
        # do incremental teleporting
        return self._rotate(
            "right", rotateDegrees, frames, smoothAnimation, agentId=agentId
        )

    def RotateLeft(
        self, rotateDegrees=90, frames=60, smoothAnimation=True, agentId=None
    ):
        # do incremental teleporting
        return self._rotate(
            "left", rotateDegrees, frames, smoothAnimation, agentId=agentId
        )

    def OrbitCameraAnimation(
        self,
        centerX,
        centerZ,
        posY,
        dx=6,
        dz=6,
        xAngle=55,
        frames=60,
        orbit_degrees_per_frame=0.5,
    ):
        """Orbits the camera around the scene.

        Example: https://www.youtube.com/watch?v=KcELPpdN770&feature=youtu.be&t=14"""
        degrees = frames * orbit_degrees_per_frame
        rot0 = self.last_event.metadata["thirdPartyCameras"][0]["rotation"][
            "y"
        ]  # starting angle
        for frame in range(frames):
            yAngle = rot0 + degrees * (frame + 1) / frames
            yield self.step(
                action="UpdateThirdPartyCamera",
                thirdPartyCameraId=0,
                rotation={"x": xAngle, "y": yAngle, "z": 0},
                position={
                    "x": centerX - dx * math.sin(math.radians(yAngle)),
                    "y": posY,
                    "z": centerZ - dz * math.cos(math.radians(yAngle)),
                },
            )

    def RelativeCameraAnimation(self, px=0, py=0, pz=0, rx=0, ry=0, rz=0, frames=60):
        """Linear interpolation between the current camera position and rotation
           and the final camera position, given by deltas to the current values.

        Params
        - px (int)=0: x offset from the current camera position.
        - py (int)=0: y offset from the current camera position.
        - pz (int)=0: z offset from the current camera position.
        - rx (int)=0: x offset from the current camera rotation.
        - ry (int)=0: y offset from the current camera rotation.
        - rz (int)=0: z offset from the current camera rotation.
        - frames (int)=60: The duration of the animation.
          Note: videos are typically 30fps."""
        for _ in range(frames):
            cam = self.last_event.metadata["thirdPartyCameras"][0]
            pos, rot = cam["position"], cam["rotation"]
            yield self.step(
                action="UpdateThirdPartyCamera",
                thirdPartyCameraId=0,
                rotation={
                    "x": rot["x"] + rx / frames,
                    "y": rot["y"] + ry / frames,
                    "z": rot["z"] + rz / frames,
                },
                position={
                    "x": pos["x"] + px / frames,
                    "y": pos["y"] + py / frames,
                    "z": pos["z"] + pz / frames,
                },
            )

    def AbsoluteCameraAnimation(
        self,
        px,
        py,
        pz,
        rx,
        ry,
        rz,
        frames=60,
        smartSkybox=True,
        FOVstart=None,
        FOVend=None,
        visibleAgents=True,
    ):
        cam = self.last_event.metadata["thirdPartyCameras"][0]
        p0, r0 = cam["position"], cam["rotation"]

        if smartSkybox:
            # toggles on and off (to give the same final result) to find the height of the ceiling
            event0 = self.step(action="ToggleMapView")
            event1 = self.step(action="ToggleMapView")
            if event0.metadata["actionReturn"]:
                maxY = event0.metadata["actionReturn"]["y"]
            else:
                maxY = event1.metadata["actionReturn"]["y"]

        for i in range(1, frames + 1):
            if self.ceiling_off and maxY > p0["y"] + (py - p0["y"]) / frames * i:
                # turn ceiling on
                self.ToggleCeiling()

            kwargs = {
                "action": "UpdateThirdPartyCamera",
                "thirdPartyCameraId": 0,
                "rotation": {
                    "x": r0["x"] + (rx - r0["x"]) / frames * i,
                    "y": r0["y"] + (ry - r0["y"]) / frames * i,
                    "z": r0["z"] + (rz - r0["z"]) / frames * i,
                },
                "position": {
                    "x": p0["x"] + (px - p0["x"]) / frames * i,
                    "y": p0["y"] + (py - p0["y"]) / frames * i,
                    "z": p0["z"] + (pz - p0["z"]) / frames * i,
                },
            }

            # enables linear animation changes to the camera FOV
            if FOVstart is not None and FOVend is not None:
                kwargs["fieldOfView"] = FOVstart + (FOVend - FOVstart) / frames * i

            if not (smartSkybox and maxY > p0["y"] + (py - p0["y"]) / frames * i):
                kwargs["skyboxColor"] = "black"
            yield self.step(**kwargs)

    def LookUp(self):
        raise NotImplementedError()

    def LookDown(self):
        raise NotImplementedError()

    def Stand(self):
        """Note: have not found an easy way to move the agent in-between
        stand and crouch."""
        raise NotImplementedError()

    def Crouch(self):
        """Note: have not found an easy way to move the agent in-between
        stand and crouch."""
        raise NotImplementedError()

    def export_video(self, path):
        """Merges all the saved frames into a .mp4 video and saves it to `path`"""
        if self.saved_frames:
            path = path if path[:-4] == ".mp4" else path + ".mp4"
            if os.path.exists(path):
                os.remove(path)
            video = cv2.VideoWriter(
                path,
                cv2.VideoWriter_fourcc(*"DIVX"),
                30,
                (self.saved_frames[0].shape[1], self.saved_frames[0].shape[0]),
            )
            for frame in self.saved_frames:
                # assumes that the frames are RGB images. CV2 uses BGR.
                video.write(cv2.cvtColor(frame, cv2.COLOR_BGR2RGB))
            cv2.destroyAllWindows()
            video.release()

    def export_frames(self, path, file_type=".png"):
        """Exports all of the presently frames to the `path` directory.

        The frames are numbered in sequential order (starting with 0)."""
        for i in range(len(self.saved_frames)):
            p = os.path.join(path, f"{i}.{file_type}")
            if os.path.exists(p):
                os.remove(p)
            Image.fromarray(self.saved_frames[i]).save(p)

    def merge_video(self, other_video_path):
        """Concatenates the frames of `other_video_path` to the presently
        generated video within this class."""
        vidcap = cv2.VideoCapture(other_video_path)
        success, image = vidcap.read()
        i = 0
        while success:
            if i % 2 == 0:
                rgb = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
                self.saved_frames.append(rgb)
            success, image = vidcap.read()
            i += 1



import numpy as np
import math


class TrialMetric(object):
    def init_trials(self, num_trials, metadata):
        ...

    def update_with_trial(self, trial_index, metadata):
        ...


class ObjectPositionVarianceAverage(TrialMetric):
    """
    Metric that computes the average of the variance of all objects in a scene across multiple runs.
    """

    def __init__(self):
        self.trials = []
        self.object_ids = []

    def init_trials(self, num_trials, metadata):
        objects = metadata["objects"]
        self.object_ids = sorted([o["objectId"] for o in objects])
        num_objects = len(self.object_ids)
        self.trials = np.empty([num_trials, num_objects, 3])

    def update_with_trial(self, trial_index, metadata):
        objects = metadata["objects"]
        object_pos_map = {
            o["objectId"]: vec_to_np_array(o["position"]) for o in objects
        }
        for object_index in range(len(self.object_ids)):
            object_id = self.object_ids[object_index]
            self.trials[trial_index][object_index] = object_pos_map[object_id]

    def compute(self, n=None):
        return np.mean(np.var(self.trials[:n], axis=0))


def vec_to_np_array(vec):
    return np.array([vec["x"], vec["y"], vec["z"]])


def trial_runner(controller, number, metric, compute_running_metric=False):
    """
    Generator that wraps metric capture from controller metadata for a number of trials
    :param controller: ai2thor controller
    :param number: int number of trials to collect metrics from
    :param metric: TrialMetric the metric to use
    :param compute_running_metric: bool whether or not to compute the metric after every trial
    :return: tuple(controller, float) with the controller and the metric after every trial
    """

    metric.init_trials(number, controller.last_event.metadata)

    for trial_index in range(number):
        try:
            yield controller, metric.compute(
                n=trial_index
            ) if compute_running_metric else math.nan
            metric.update_with_trial(trial_index, controller.last_event.metadata)
            controller.reset()
        except RuntimeError as e:
            print(
                e,
                "Last action status: {}".format(
                    controller.last_event.meatadata["actionSuccess"]
                ),
                controller.last_event.meatadata["errorMessage"],
            )
    yield controller, metric.compute()



import subprocess
import os.path
import re
import tempfile
from ai2thor.build import COMMIT_ID
import shlex
import shutil


def pci_records():
    records = []
    command = shlex.split("lspci -vmm")
    output = subprocess.check_output(command).decode()

    for devices in output.strip().split("\n\n"):
        record = {}
        records.append(record)
        for row in devices.split("\n"):
            key, value = row.split("\t")
            record[key.split(":")[0]] = value

    return records


def xorg_bus_id():
    bus_id = None
    for r in pci_records():
        if r.get("Vendor", "") == "NVIDIA Corporation" and r["Class"] in [
            "VGA compatible controller",
            "3D controller",
        ]:
            bus_id = "PCI:" + ":".join(
                map(lambda x: str(int(x, 16)), re.split(r"[:\.]", r["Slot"]))
            )
            break

    if bus_id is None:
        raise Exception("no valid nvidia device could be found")

    return bus_id


def has_docker():
    with open(os.devnull, "w") as dn:
        return subprocess.call(["which", "docker"], stdout=dn) == 0


def bridge_gateway():
    output = (
        subprocess.check_output(
            "docker network inspect -f '{{range .IPAM.Config}}{{.Gateway}}{{end}}' bridge",
            shell=True,
        )
        .decode("ascii")
        .strip()
    )
    if re.match(r"^[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+$", output):
        return output
    else:
        raise Exception(
            "Didn't receive a single ip address from network inspect bridge: %s"
            % output
        )


def nvidia_version():
    version = None
    version_path = "/proc/driver/nvidia/version"
    if os.path.isfile(version_path):
        with open(version_path) as f:
            for line in f:
                if line.startswith("NVRM version: NVIDIA"):
                    match = re.search(r"Kernel Module\s+([0-9\.]+)\s+", line)
                    if match:
                        version = match.group(1)
                        break
    return version


def generate_dockerfile(tag):

    driver_url = "http://us.download.nvidia.com/XFree86/Linux-x86_64/{version}/NVIDIA-Linux-x86_64-{version}.run".format(
        version=nvidia_version()
    )
    driver_filename = os.path.basename(driver_url)

    dockerfile = """
FROM ai2thor/ai2thor-base:{tag}
RUN wget {driver_url} -P /root/
RUN sh /root/{driver_filename} -s --no-kernel-module
""".format(
        driver_filename=driver_filename, driver_url=driver_url, tag=tag
    )

    return dockerfile


def image_exists(image_name):
    output = subprocess.check_output("docker images -q %s" % image_name, shell=True)
    return len(output) > 0


def run(image_name, base_dir, command, environment):
    allowed_keys = {"AI2THOR_PORT", "AI2THOR_CLIENT_TOKEN", "AI2THOR_HOST"}

    environment_string = ""
    for k, v in environment.items():
        if k in allowed_keys:
            environment_string += " -e %s=%s " % (k, v)

    relative_command = os.path.relpath(command, base_dir)
    docker_command = os.path.join("/root/.ai2thor", relative_command)
    environment_string += " -e %s=%s -e %s='%s'" % (
        "AI2THOR_DEVICE_BUSID",
        xorg_bus_id(),
        "AI2THOR_COMMAND",
        docker_command,
    )
    command = "docker run -v {base_dir}:/root/.ai2thor -d --privileged {environment} {image_name} /root/start.sh".format(
        environment=environment_string, image_name=image_name, base_dir=base_dir
    )
    container_id = subprocess.check_output(command, shell=True).decode("ascii").strip()
    return container_id


def kill_container(container_id):
    subprocess.check_output("docker kill %s" % container_id, shell=True)


def build_image():

    version = nvidia_version()
    image_name = "ai2thor/ai2thor-nvidia-%s:%s" % (version, COMMIT_ID)

    if image_exists(image_name):
        return image_name

    td = tempfile.mkdtemp()
    with open("%s/Dockerfile" % td, "w") as f:
        f.write(generate_dockerfile(tag))

    subprocess.check_call("docker build --rm -t %s %s" % (image_name, td), shell=True)

    shutil.rmtree(td)

    return image_name



import os
import time
import msgpack
import numpy as np
import requests
import cv2
from pprint import pprint
import shutil
import copy

from ai2thor._quality_settings import QUALITY_SETTINGS, DEFAULT_QUALITY
from ai2thor.server import Event, MultiAgentEvent, DepthFormat
from ai2thor.interact import InteractiveControllerPrompt, DefaultActions


class Controller(object):
    def __init__(
        self,
        headless=False,
        host="127.0.0.1",
        port=0,
        width=300,
        height=300,
        agent_id=0,
        image_dir=".",
        save_image_per_frame=False,
        depth_format=DepthFormat.Meters,
        camera_near_plane=0.1,
        camera_far_plane=20,
    ):
        self.host = host
        self.port = port
        self.headless = headless
        self.last_event = {}
        self.last_action = {}
        self.sequence_id = 0
        self.agent_id = agent_id
        self.screen_width = width
        self.screen_height = height
        self.depth_format = depth_format
        self.camera_near_plane = (camera_near_plane,)
        self.camera_far_plane = camera_far_plane

        if image_dir != ".":
            if os.path.exists(image_dir):
                shutil.rmtree(image_dir)
            os.makedirs(image_dir)

        self.interactive_controller = InteractiveControllerPrompt(
            [
                DefaultActions.MoveAhead,
                DefaultActions.MoveBack,
                DefaultActions.RotateLeft,
                DefaultActions.RotateRight,
                DefaultActions.LookUp,
                DefaultActions.LookDown,
            ],
            image_dir=image_dir,
            image_per_frame=save_image_per_frame,
        )

        self.start(port, host, agent_id=agent_id)

    def stop(self):
        pass

    def start(self, port=9200, host="127.0.0.1", agent_id=0, **kwargs):
        self.host = host
        self.port = port
        self.agent_id = agent_id

        # response_payload = self._post_event('start')
        pprint("-- Start:")
        # pprint(response_payload)

    def reset(self, scene_name=None):
        if not scene_name.endswith("_physics"):
            scene_name = scene_name + "_physics"
        self.sequence_id = 0
        # response_payload = self._post_event(
        #     'reset', dict(action='Reset', sceneName=scene_name, sequenceId=self.sequence_id)
        # )
        pprint("-- Reset:")
        # pprint(response_payload)

        return self.last_event

    def step(self, action=None, **action_args):

        if type(action) is dict:
            action = copy.deepcopy(action)  # prevent changes from leaking
        else:
            action = dict(action=action)

        raise_for_failure = action_args.pop("raise_for_failure", False)

        action.update(action_args)

        if self.headless:
            action["renderImage"] = False

        action["sequenceId"] = self.sequence_id
        action["agentId"] = self.agent_id

        self.last_action = action

        rotation = action.get("rotation")
        if rotation is not None and type(rotation) != dict:
            action["rotation"] = {}
            action["rotation"]["y"] = rotation

        payload = self._post_event("step", action)
        events = []
        for i, agent_metadata in enumerate(payload["metadata"]["agents"]):
            event = Event(agent_metadata)
            image_mapping = dict(
                image=lambda x: event.add_image(x, flip_y=False, flip_rb_colors=False),
                image_depth=lambda x: event.add_image_depth_robot(
                    x,
                    self.depth_format,
                    camera_near_plane=self.camera_near_plane,
                    camera_far_plane=self.camera_far_plane,
                    flip_y=False,
                    dtype=np.float64,
                ),
            )
            for key in image_mapping.keys():
                if key in payload and len(payload[key]) > i:
                    image_mapping[key](payload[key][i])
            events.append(event)

        if len(events) > 1:
            self.last_event = MultiAgentEvent(self.agent_id, events)
        else:
            self.last_event = events[0]

        if (
            not self.last_event.metadata["lastActionSuccess"]
            and self.last_event.metadata["errorCode"] == "InvalidAction"
        ):
            raise ValueError(self.last_event.metadata["errorMessage"])

        if raise_for_failure:
            assert self.last_event.metadata["lastActionSuccess"]

        # pprint("Display event:")
        # Controller._display_step_event(self.last_event)

        return self.last_event

    def interact(
        self,
        semantic_segmentation_frame=False,
        instance_segmentation_frame=False,
        depth_frame=False,
        color_frame=False,
        metadata=False,
    ):
        self.interactive_controller.interact(
            self,
            semantic_segmentation_frame=semantic_segmentation_frame,
            instance_segmentation_frame=instance_segmentation_frame,
            depth_frame=depth_frame,
            color_frame=color_frame,
            metadata=metadata,
        )

    def _post_event(self, route="", data=None):
        r = requests.post(self._get_url(route), json=data)
        s = time.time()
        pprint('ACTION "{}"'.format(data["action"]))
        pprint("POST")
        # pprint(r.content)
        pprint(r.status_code)
        return msgpack.unpackb(r.content, raw=False)

    def _get_url(self, route=""):
        return "http://{}:{}{}".format(
            self.host, self.port, "/{}".format(route) if route != "" else ""
        )

    @staticmethod
    def _display_step_event(event):
        metadata = event.metadata
        image = event.frame
        pprint(metadata)
        cv2.imshow("aoeu", event.cv2img)
        cv2.waitKey(1000)



from aws_requests_auth.boto_utils import BotoAWSRequestsAuth
import os
import requests
import json
from ai2thor.util import makedirs
import time
import random
import ai2thor.downloader
import zipfile
import logging
from ai2thor.util.lock import LockSh, LockEx
from ai2thor.util import atomic_write
import io

logger = logging.getLogger(__name__)

PUBLIC_S3_BUCKET = "ai2-thor-public"
PUBLIC_WEBGL_S3_BUCKET = "ai2-thor-webgl-public"
PRIVATE_S3_BUCKET = "ai2-thor-private"
PYPI_S3_BUCKET = "ai2-thor-pypi"

COMMIT_ID = None
try:
    from ai2thor._builds import COMMIT_ID
except ImportError:
    pass

platform_map = dict(Linux64="Linux", OSXIntel64="Darwin")
arch_platform_map = {v: k for k, v in platform_map.items()}


def build_name(arch, commit_id, include_private_scenes=False):
    if include_private_scenes:
        return "thor-private-%s-%s" % (arch, commit_id)
    else:
        return "thor-%s-%s" % (arch, commit_id)


def boto_auth():
    return BotoAWSRequestsAuth(
        aws_host="s3-us-west-2.amazonaws.com", aws_region="us-west-2", aws_service="s3"
    )


base_url = "http://s3-us-west-2.amazonaws.com/%s/" % PUBLIC_S3_BUCKET
private_base_url = "http://s3-us-west-2.amazonaws.com/%s/" % PRIVATE_S3_BUCKET


# dummy build when connecting to the editor
class EditorBuild(object):
    def __init__(self):
        # assuming that an external build supports both server types
        self.server_types = ["FIFO", "WSGI"]
        self.url = None
        self.unity_proc = None

    def download(self):
        pass

    def unlock(self):
        pass

    def lock_sh(self):
        pass


class ExternalBuild(object):
    def __init__(self, executable_path):
        self.executable_path = executable_path

        # assuming that an external build supports both server types
        self.server_types = ["FIFO", "WSGI"]

    def download(self):
        pass

    def unlock(self):
        pass

    def lock_sh(self):
        pass


class Build(object):
    def __init__(self, arch, commit_id, include_private_scenes, releases_dir=None):

        self.arch = arch
        self.commit_id = commit_id
        self.include_private_scenes = include_private_scenes
        self.releases_dir = releases_dir
        self.tmp_dir = None

        if self.releases_dir:
            self.tmp_dir = os.path.normpath(self.releases_dir + "/../tmp")

    def download(self):

        makedirs(self.releases_dir)
        makedirs(self.tmp_dir)

        with LockEx(os.path.join(self.tmp_dir, self.name)):
            if not os.path.isdir(self.base_dir):
                z = self.zipfile()
                # use tmpdir instead or a random number
                extract_dir = os.path.join(self.tmp_dir, self.name)
                logger.debug("Extracting zipfile %s" % os.path.basename(self.url))
                z.extractall(extract_dir)

                os.rename(extract_dir, self.base_dir)
                # This can be removed after migrating OSXIntel64 builds to have the AI2Thor executable
                if os.path.exists(self.old_executable_path) and not os.path.exists(
                    self.executable_path
                ):
                    os.link(self.old_executable_path, self.executable_path)

                # we can lose the executable permission when unzipping a build
                os.chmod(self.executable_path, 0o755)

            else:
                logger.debug("%s exists - skipping download" % (self.executable_path,))

    def zipfile(self):
        zip_data = ai2thor.downloader.download(
            self.url, self.sha256(), self.include_private_scenes
        )

        return zipfile.ZipFile(io.BytesIO(zip_data))

    def download_metadata(self):
        # this can happen if someone has an existing release without the metadata
        # built prior to the backfill
        # can add check to see if metadata has expired/we update metadata
        # if we want to add more info to metadata
        if not self.metadata:
            z = self.zipfile()
            atomic_write(self.metadata_path, z.read("metadata.json"))

    @property
    def old_executable_path(self):
        if self.arch == "Linux64":
            return self.executable_path
        elif self.arch == "OSXIntel64":
            return os.path.join(
                self.base_dir, self.name + ".app", "Contents/MacOS", self.name
            )
        else:
            raise Exception("unable to handle target arch %s" % self.arch)

    def parse_plist(self):
        import xml.etree.ElementTree as ET

        plist_path = os.path.join(
            self.base_dir, self.name + ".app", "Contents/Info.plist"
        )

        with open(plist_path) as f:
            plist = f.read()
        root = ET.fromstring(plist)

        keys = [x.text for x in root.findall("dict/key")]
        values = [x.text for x in root.findall("dict/string")]
        return dict(zip(keys, values))

    @property
    def base_dir(self):
        return os.path.join(self.releases_dir, self.name)

    @property
    def executable_path(self):
        if self.arch == "Linux64":
            return os.path.join(self.base_dir, self.name)
        elif self.arch == "OSXIntel64":
            plist = self.parse_plist()
            return os.path.join(
                self.base_dir,
                self.name + ".app",
                "Contents/MacOS",
                plist["CFBundleExecutable"],
            )
        else:
            raise Exception("unable to handle target arch %s" % self.arch)

    @property
    def metadata_path(self):
        return os.path.join(self.base_dir, "metadata.json")

    @property
    def metadata(self):
        if os.path.isfile(self.metadata_path):
            with open(self.metadata_path, "r") as f:
                return json.loads(f.read())
        else:
            return None

    @property
    def server_types(self):
        self.download_metadata()
        return self.metadata.get("server_types", [])

    def auth(self):
        if self.include_private_scenes:
            return boto_auth()
        else:
            return None

    def _base_url(self):
        if self.include_private_scenes:
            return private_base_url
        else:
            return base_url

    @property
    def url(self):
        return self._base_url() + os.path.join("builds", self.name + ".zip")

    @property
    def name(self):
        return build_name(self.arch, self.commit_id, self.include_private_scenes)

    def unlock(self):
        if self._lock:
            self._lock.unlock()
            self._lock = None

    def lock_sh(self):
        self._lock = LockSh(self.base_dir)
        self._lock.lock()

    @property
    def log_url(self):
        return os.path.splitext(self.url)[0] + ".log"

    @property
    def metadata_url(self):
        return os.path.splitext(self.url)[0] + ".json"

    @property
    def sha256_url(self):
        return os.path.splitext(self.url)[0] + ".sha256"

    def exists(self):
        return (self.releases_dir and os.path.isdir(self.base_dir)) or (
            requests.head(self.url, auth=self.auth()).status_code == 200
        )

    def log_exists(self):
        return requests.head(self.log_url, auth=self.auth()).status_code == 200

    def sha256(self):
        res = requests.get(self.sha256_url, auth=self.auth())
        res.raise_for_status()
        return res.content.decode("ascii")



# GENERATED FILE - DO NOT EDIT
DEFAULT_QUALITY = 'Ultra'
QUALITY_SETTINGS = {'DONOTUSE': 0,
 'High': 5,
 'High WebGL': 8,
 'Low': 2,
 'Medium': 3,
 'MediumCloseFitShadows': 4,
 'Ultra': 7,
 'Very High': 6,
 'Very Low': 1}


import sys
from PIL import Image
import numpy as np
import os

from enum import Enum
import json


class DefaultActions(Enum):
    MoveRight = (0,)
    MoveLeft = (1,)
    MoveAhead = (2,)
    MoveBack = (3,)
    LookUp = (4,)
    LookDown = (5,)
    RotateRight = (8,)
    RotateLeft = 9


# TODO tie this with actions
# class ObjectActions(Enum):
#     PutObject
#     MoveHandAhead
#     MoveHandBack
#     MoveHandRight
#     MoveHandLeft
#     MoveHandUp
#     MoveHandDown
#     DropHandObject
#     PickupObject,
#     OpenObject,
#     CloseObject,
#     ToggleObjectOff


def get_term_character():
    # NOTE: Leave these imports here! They are incompatible with Windows.
    import tty
    import termios

    fd = sys.stdin.fileno()
    old_settings = termios.tcgetattr(fd)
    try:
        tty.setraw(sys.stdin.fileno())
        ch = sys.stdin.read(1)
    finally:
        termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
    return ch


class InteractiveControllerPrompt(object):
    def __init__(
        self,
        default_actions,
        has_object_actions=True,
        image_dir=".",
        image_per_frame=False,
    ):
        self.default_actions = default_actions
        self.has_object_actions = has_object_actions
        self.image_per_frame = image_per_frame
        self.image_dir = image_dir
        self.counter = 0

        default_interact_commands = {
            "\x1b[C": dict(action="MoveRight", moveMagnitude=0.25),
            "\x1b[D": dict(action="MoveLeft", moveMagnitude=0.25),
            "\x1b[A": dict(action="MoveAhead", moveMagnitude=0.25),
            "\x1b[B": dict(action="MoveBack", moveMagnitude=0.25),
            "\x1b[1;2A": dict(action="LookUp"),
            "\x1b[1;2B": dict(action="LookDown"),
            "i": dict(action="LookUp"),
            "k": dict(action="LookDown"),
            "l": dict(action="RotateRight"),
            "j": dict(action="RotateLeft"),
            "\x1b[1;2C": dict(action="RotateRight"),
            "\x1b[1;2D": dict(action="RotateLeft"),
        }
        action_set = {a.name for a in default_actions}

        self.default_interact_commands = {
            k: v
            for (k, v) in default_interact_commands.items()
            if v["action"] in action_set
        }

    def interact(
        self,
        controller,
        semantic_segmentation_frame=False,
        instance_segmentation_frame=False,
        depth_frame=False,
        color_frame=False,
        metadata=False,
    ):

        if not sys.stdout.isatty():
            raise RuntimeError("controller.interact() must be run from a terminal")

        default_interact_commands = self.default_interact_commands

        self._interact_commands = default_interact_commands.copy()

        command_message = u"Enter a Command: Move \u2190\u2191\u2192\u2193, Rotate/Look Shift + \u2190\u2191\u2192\u2193, Quit 'q' or Ctrl-C"
        print(command_message)
        for a in self.next_interact_command():
            new_commands = {}
            command_counter = dict(counter=1)

            def add_command(cc, action, **args):
                if cc["counter"] < 15:
                    com = dict(action=action)
                    com.update(args)
                    new_commands[str(cc["counter"])] = com
                    cc["counter"] += 1

            event = controller.step(a)
            visible_objects = []
            InteractiveControllerPrompt.write_image(
                event,
                self.image_dir,
                "_{}".format(self.counter),
                image_per_frame=self.image_per_frame,
                semantic_segmentation_frame=semantic_segmentation_frame,
                instance_segmentation_frame=instance_segmentation_frame,
                color_frame=color_frame,
                depth_frame=depth_frame,
                metadata=metadata,
            )

            self.counter += 1
            if self.has_object_actions:
                for o in event.metadata["objects"]:
                    if o["visible"]:
                        visible_objects.append(o["objectId"])
                        if o["openable"]:
                            if o["isOpen"]:
                                add_command(
                                    command_counter,
                                    "CloseObject",
                                    objectId=o["objectId"],
                                )
                            else:
                                add_command(
                                    command_counter,
                                    "OpenObject",
                                    objectId=o["objectId"],
                                )

                        if o["toggleable"]:
                            add_command(
                                command_counter,
                                "ToggleObjectOff",
                                objectId=o["objectId"],
                            )

                        if len(event.metadata["inventoryObjects"]) > 0:
                            inventoryObjectId = event.metadata["inventoryObjects"][0][
                                "objectId"
                            ]
                            if (
                                o["receptacle"]
                                and (not o["openable"] or o["isOpen"])
                                and inventoryObjectId != o["objectId"]
                            ):
                                add_command(
                                    command_counter,
                                    "PutObject",
                                    objectId=inventoryObjectId,
                                    receptacleObjectId=o["objectId"],
                                )
                                add_command(
                                    command_counter, "MoveHandAhead", moveMagnitude=0.1
                                )
                                add_command(
                                    command_counter, "MoveHandBack", moveMagnitude=0.1
                                )
                                add_command(
                                    command_counter, "MoveHandRight", moveMagnitude=0.1
                                )
                                add_command(
                                    command_counter, "MoveHandLeft", moveMagnitude=0.1
                                )
                                add_command(
                                    command_counter, "MoveHandUp", moveMagnitude=0.1
                                )
                                add_command(
                                    command_counter, "MoveHandDown", moveMagnitude=0.1
                                )
                                add_command(command_counter, "DropHandObject")

                        elif o["pickupable"]:
                            add_command(
                                command_counter, "PickupObject", objectId=o["objectId"]
                            )

            self._interact_commands = default_interact_commands.copy()
            self._interact_commands.update(new_commands)

            print("Position: {}".format(event.metadata["agent"]["position"]))
            print(command_message)
            print("Visible Objects:\n" + "\n".join(sorted(visible_objects)))

            skip_keys = ["action", "objectId"]
            for k in sorted(new_commands.keys()):
                v = new_commands[k]
                command_info = [k + ")", v["action"]]
                if "objectId" in v:
                    command_info.append(v["objectId"])

                for a, av in v.items():
                    if a in skip_keys:
                        continue
                    command_info.append("%s: %s" % (a, av))

                print(" ".join(command_info))

    def next_interact_command(self):

        current_buffer = ""
        while True:
            commands = self._interact_commands
            current_buffer += get_term_character()
            if current_buffer == "q" or current_buffer == "\x03":
                break

            if current_buffer in commands:
                yield commands[current_buffer]
                current_buffer = ""
            else:
                match = False
                for k, v in commands.items():
                    if k.startswith(current_buffer):
                        match = True
                        break

                if not match:
                    current_buffer = ""

    @classmethod
    def write_image(
        cls,
        event,
        image_dir,
        suffix,
        image_per_frame=False,
        semantic_segmentation_frame=False,
        instance_segmentation_frame=False,
        depth_frame=False,
        color_frame=False,
        metadata=False,
    ):
        def save_image(name, image, flip_br=False):
            # TODO try to use PIL which did not work with RGBA
            # image.save(
            #     name
            # )
            import cv2

            img = image
            if flip_br:
                img = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
            cv2.imwrite("{}.png".format(name), img)

        def array_to_image(arr, mode=None):
            return arr

        def json_write(name, obj):
            with open("{}.json".format(name), "w") as outfile:
                json.dump(obj, outfile, indent=4, sort_keys=True)

        frame_writes = [
            (
                "color",
                color_frame,
                lambda event: event.frame,
                array_to_image,
                lambda x, y: save_image(x, y, flip_br=True),
            ),
            (
                "instance_segmentation",
                instance_segmentation_frame,
                lambda event: event.instance_segmentation_frame,
                array_to_image,
                save_image,
            ),
            (
                "class_segmentation",
                semantic_segmentation_frame,
                lambda event: event.semantic_segmentation_frame,
                array_to_image,
                save_image,
            ),
            (
                "depth",
                depth_frame,
                lambda event: event.depth_frame,
                lambda data: array_to_image(
                    (255.0 / data.max() * (data - data.min())).astype(np.uint8)
                ),
                save_image,
            ),
            (
                "depth_raw",
                depth_frame,
                lambda event: event.depth_frame,
                lambda x: x,
                lambda name, x: np.save(
                    name.strip(".png").strip("./")
                    if image_dir == "."
                    else name.strip(".png"),
                    x.astype(np.float32),
                ),
            ),
            (
                "metadata",
                metadata,
                lambda event: event.metadata,
                lambda x: x,
                json_write,
            ),
        ]

        for frame_filename, condition, frame_func, transform, save in frame_writes:
            frame = frame_func(event)
            if frame is not None:
                frame = transform(frame)
                image_name = os.path.join(
                    image_dir,
                    "{}{}".format(
                        frame_filename, "{}".format(suffix) if image_per_frame else ""
                    ),
                )
                print("Image {}, {}".format(image_name, image_dir))
                save(image_name, frame)

            else:
                print("No frame present, call initialize with the right parameters")



from flask import Flask, request, jsonify, make_response
import numpy as np
import io
import threading
import random
import msgpack
import werkzeug
import werkzeug.serving
import werkzeug.http
import os


def random_image(height, width):
    img = np.zeros(height * width * 3, dtype=np.uint8).reshape(height, width, 3)
    img[:, :, 0] = random.randint(0, 255)
    img[:, :, 1] = random.randint(0, 255)
    img[:, :, 2] = random.randint(0, 255)

    return img


class MockServer(object):
    def __init__(self, height, width):
        app = Flask(__name__)

        self.height = height
        self.width = width

        @app.route("/ping", methods=["GET"])
        def ping():
            return make_response("PONG")

        @app.route("/start", methods=["POST"])
        def start():
            return make_response(msgpack.packb({"status": 200}, use_bin_type=True))

        @app.route("/reset", methods=["POST"])
        def reset():
            return make_response(msgpack.packb({"status": 200}, use_bin_type=True))

        @app.route("/step", methods=["POST"])
        def step():
            content = request.json

            metadata = {
                u"sequenceId": content["sequenceId"],
                u"agents": [
                    {
                        u"agentId": 0,
                        u"screenHeight": self.height,
                        u"screenWidth": self.width,
                        u"lastAction": content["action"],
                        u"lastActionSuccess": True,
                    }
                ],
            }

            result = {
                u"image": [random_image(self.height, self.width).tostring()],
                u"image_depth": [],
                u"metadata": metadata,
            }
            out = msgpack.packb(result, use_bin_type=True)

            return make_response(out)

        self.host = "127.0.0.1"
        self.port = 9200
        app.config.update(PROPAGATE_EXCEPTIONS=True, JSONIFY_PRETTYPRINT_REGULAR=False)
        self.wsgi_server = werkzeug.serving.make_server(self.host, self.port, app)

    def start(self):
        self.wsgi_server.serve_forever()



# Copyright Allen Institute for Artificial Intelligence 2017
"""
ai2thor.controller

Primary entrypoint into the Thor API. Provides all the high-level functions
needed to control the in-game agent through ai2thor.server.

"""
import atexit
from collections import deque, defaultdict
from itertools import product
import json
import copy
import logging
import math
import time
import random
import shlex
import signal
import subprocess
import shutil
import re
import os
import platform
import uuid
import sys
from functools import lru_cache


import numpy as np
import ai2thor.wsgi_server
import ai2thor.fifo_server
from ai2thor.exceptions import UnityCrashException, RestartError
from ai2thor.interact import InteractiveControllerPrompt, DefaultActions
from ai2thor.server import DepthFormat
import ai2thor.build
from ai2thor._quality_settings import QUALITY_SETTINGS, DEFAULT_QUALITY
from ai2thor.util import makedirs, atomic_write
from ai2thor.util.lock import LockEx

import warnings

logger = logging.getLogger(__name__)

RECEPTACLE_OBJECTS = {
    "Box": {
        "Candle",
        "CellPhone",
        "Cloth",
        "CreditCard",
        "Dirt",
        "KeyChain",
        "Newspaper",
        "ScrubBrush",
        "SoapBar",
        "SoapBottle",
        "ToiletPaper",
    },
    "Cabinet": {
        "Bowl",
        "BowlDirty",
        "Box",
        "Bread",
        "BreadSliced",
        "ButterKnife",
        "Candle",
        "CellPhone",
        "Cloth",
        "CoffeeMachine",
        "Container",
        "ContainerFull",
        "CreditCard",
        "Cup",
        "Fork",
        "KeyChain",
        "Knife",
        "Laptop",
        "Mug",
        "Newspaper",
        "Pan",
        "Plate",
        "Plunger",
        "Pot",
        "Potato",
        "Sandwich",
        "ScrubBrush",
        "SoapBar",
        "SoapBottle",
        "Spoon",
        "SprayBottle",
        "Statue",
        "TissueBox",
        "Toaster",
        "ToiletPaper",
        "WateringCan",
    },
    "CoffeeMachine": {"MugFilled", "Mug"},
    "CounterTop": {
        "Apple",
        "AppleSlice",
        "Bowl",
        "BowlDirty",
        "BowlFilled",
        "Box",
        "Bread",
        "BreadSliced",
        "ButterKnife",
        "Candle",
        "CellPhone",
        "CoffeeMachine",
        "Container",
        "ContainerFull",
        "CreditCard",
        "Cup",
        "Egg",
        "EggFried",
        "EggShell",
        "Fork",
        "HousePlant",
        "KeyChain",
        "Knife",
        "Laptop",
        "Lettuce",
        "LettuceSliced",
        "Microwave",
        "Mug",
        "MugFilled",
        "Newspaper",
        "Omelette",
        "Pan",
        "Plate",
        "Plunger",
        "Pot",
        "Potato",
        "PotatoSliced",
        "RemoteControl",
        "Sandwich",
        "ScrubBrush",
        "SoapBar",
        "SoapBottle",
        "Spoon",
        "SprayBottle",
        "Statue",
        "Television",
        "TissueBox",
        "Toaster",
        "ToiletPaper",
        "Tomato",
        "TomatoSliced",
        "WateringCan",
    },
    "Fridge": {
        "Apple",
        "AppleSlice",
        "Bowl",
        "BowlDirty",
        "BowlFilled",
        "Bread",
        "BreadSliced",
        "Container",
        "ContainerFull",
        "Cup",
        "Egg",
        "EggFried",
        "EggShell",
        "Lettuce",
        "LettuceSliced",
        "Mug",
        "MugFilled",
        "Omelette",
        "Pan",
        "Plate",
        "Pot",
        "Potato",
        "PotatoSliced",
        "Sandwich",
        "Tomato",
        "TomatoSliced",
    },
    "GarbageCan": {
        "Apple",
        "AppleSlice",
        "Box",
        "Bread",
        "BreadSliced",
        "Candle",
        "CellPhone",
        "CreditCard",
        "Egg",
        "EggFried",
        "EggShell",
        "LettuceSliced",
        "Newspaper",
        "Omelette",
        "Plunger",
        "Potato",
        "PotatoSliced",
        "Sandwich",
        "ScrubBrush",
        "SoapBar",
        "SoapBottle",
        "SprayBottle",
        "Statue",
        "ToiletPaper",
        "Tomato",
        "TomatoSliced",
    },
    "Microwave": {
        "Bowl",
        "BowlDirty",
        "BowlFilled",
        "Bread",
        "BreadSliced",
        "Container",
        "ContainerFull",
        "Cup",
        "Egg",
        "EggFried",
        "Mug",
        "MugFilled",
        "Omelette",
        "Plate",
        "Potato",
        "PotatoSliced",
        "Sandwich",
    },
    "PaintingHanger": {"Painting"},
    "Pan": {
        "Apple",
        "AppleSlice",
        "EggFried",
        "Lettuce",
        "LettuceSliced",
        "Omelette",
        "Potato",
        "PotatoSliced",
        "Tomato",
        "TomatoSliced",
    },
    "Pot": {
        "Apple",
        "AppleSlice",
        "EggFried",
        "Lettuce",
        "LettuceSliced",
        "Omelette",
        "Potato",
        "PotatoSliced",
        "Tomato",
        "TomatoSliced",
    },
    "Sink": {
        "Apple",
        "AppleSlice",
        "Bowl",
        "BowlDirty",
        "BowlFilled",
        "ButterKnife",
        "Container",
        "ContainerFull",
        "Cup",
        "Egg",
        "EggFried",
        "EggShell",
        "Fork",
        "Knife",
        "Lettuce",
        "LettuceSliced",
        "Mug",
        "MugFilled",
        "Omelette",
        "Pan",
        "Plate",
        "Pot",
        "Potato",
        "PotatoSliced",
        "Sandwich",
        "ScrubBrush",
        "SoapBottle",
        "Spoon",
        "Tomato",
        "TomatoSliced",
        "WateringCan",
    },
    "StoveBurner": {"Omelette", "Pot", "Pan", "EggFried"},
    "TableTop": {
        "Apple",
        "AppleSlice",
        "Bowl",
        "BowlDirty",
        "BowlFilled",
        "Box",
        "Bread",
        "BreadSliced",
        "ButterKnife",
        "Candle",
        "CellPhone",
        "CoffeeMachine",
        "Container",
        "ContainerFull",
        "CreditCard",
        "Cup",
        "Egg",
        "EggFried",
        "EggShell",
        "Fork",
        "HousePlant",
        "KeyChain",
        "Knife",
        "Laptop",
        "Lettuce",
        "LettuceSliced",
        "Microwave",
        "Mug",
        "MugFilled",
        "Newspaper",
        "Omelette",
        "Pan",
        "Plate",
        "Plunger",
        "Pot",
        "Potato",
        "PotatoSliced",
        "RemoteControl",
        "Sandwich",
        "ScrubBrush",
        "SoapBar",
        "SoapBottle",
        "Spoon",
        "SprayBottle",
        "Statue",
        "Television",
        "TissueBox",
        "Toaster",
        "ToiletPaper",
        "Tomato",
        "TomatoSliced",
        "WateringCan",
    },
    "ToiletPaperHanger": {"ToiletPaper"},
    "TowelHolder": {"Cloth"},
}


def process_alive(pid):
    """
    Use kill(0) to determine if pid is alive
    :param pid: process id
    :rtype: bool
    """
    try:
        os.kill(pid, 0)
    except OSError:
        return False

    return True


def distance(point1, point2):
    x_diff = (point1["x"] - point2["x"]) ** 2
    z_diff = (point1["z"] - point2["z"]) ** 2
    return math.sqrt(x_diff + z_diff)


def key_for_point(x, z):
    return "%0.1f %0.1f" % (x, z)


class Controller(object):
    def __init__(
        self,
        quality=DEFAULT_QUALITY,
        fullscreen=False,
        headless=False,
        port=0,
        start_unity=True,
        local_executable_path=None,
        local_build=False,
        commit_id=ai2thor.build.COMMIT_ID,
        branch=None,
        width=300,
        height=300,
        x_display=None,
        host="127.0.0.1",
        scene=None,
        image_dir=".",
        save_image_per_frame=False,
        depth_format=DepthFormat.Meters,
        add_depth_noise=False,
        download_only=False,
        include_private_scenes=False,
        server_class=ai2thor.fifo_server.FifoServer,
        **unity_initialization_parameters,
    ):
        self.receptacle_nearest_pivot_points = {}
        self.server = None
        self.unity_pid = None
        self.container_id = None
        self.width = width
        self.height = height
        self.x_display = x_display

        self.last_event = None
        self.scene = None
        self._scenes_in_build = None
        self.killing_unity = False
        self.quality = quality
        self.lock_file = None
        self.fullscreen = fullscreen
        self.headless = headless
        self.depth_format = depth_format
        self.add_depth_noise = add_depth_noise
        self.include_private_scenes = include_private_scenes
        self.server_class = server_class
        self._build = None

        self.interactive_controller = InteractiveControllerPrompt(
            list(DefaultActions),
            has_object_actions=True,
            image_dir=image_dir,
            image_per_frame=save_image_per_frame,
        )

        if not start_unity:
            self._build = ai2thor.build.EditorBuild()
        elif local_executable_path:
            self._build = ai2thor.build.ExternalBuild(local_executable_path)
        else:
            self._build = self.find_build(local_build, commit_id, branch)

        if self._build is None:
            raise Exception(
                "Couldn't find a suitable build for platform: %s" % platform.system()
            )

        self._build.download()

        if not download_only:
            self.start(
                port=port,
                start_unity=start_unity,
                width=width,
                height=height,
                x_display=x_display,
                host=host,
            )

            self.initialization_parameters = unity_initialization_parameters

            if "continuous" in self.initialization_parameters:
                warnings.warn(
                    "Warning: 'continuous' is deprecated and will be ignored,"
                    " use 'snapToGrid={}' instead.".format(
                        not self.initialization_parameters["continuous"]
                    ),
                    DeprecationWarning,
                )

            if (
                "fastActionEmit" in self.initialization_parameters
                and self.server_class != ai2thor.fifo_server.FifoServer
            ):
                warnings.warn("fastAtionEmit is only available with the FifoServer")

            if "continuousMode" in self.initialization_parameters:
                warnings.warn(
                    "Warning: 'continuousMode' is deprecated and will be ignored,"
                    " use 'snapToGrid={}' instead.".format(
                        not self.initialization_parameters["continuousMode"]
                    ),
                    DeprecationWarning,
                )

            # Let's set the scene for them!
            if scene is None:
                scenes_in_build = self.scenes_in_build
                if not scenes_in_build:
                    raise RuntimeError("No scenes are in your build of AI2-THOR!")

                # use a robothor scene
                robothor_scenes = set(self.robothor_scenes())

                # prioritize robothor if locobot is being used
                robothor_scenes_in_build = robothor_scenes.intersection(scenes_in_build)

                # check for bot as well, for backwards compatibility support
                if (
                    unity_initialization_parameters.get("agentMode", "default").lower()
                    in {"locobot", "bot"}
                    and robothor_scenes_in_build
                ):
                    # get the first robothor scene
                    scene = sorted(list(robothor_scenes_in_build))[0]
                else:
                    ithor_scenes = set(self.ithor_scenes())
                    ithor_scenes_in_build = ithor_scenes.intersection(scenes_in_build)
                    if ithor_scenes_in_build:
                        # prioritize iTHOR because that's what the default agent best uses
                        scene = sorted(list(ithor_scenes_in_build))[0]
                    else:
                        # perhaps only using RoboTHOR or using only custom scenes
                        scene = sorted(list(scenes_in_build))[0]

            event = self.reset(scene)

            # older builds don't send actionReturn on Initialize
            init_return = event.metadata["actionReturn"]
            if init_return:
                self.server.set_init_params(init_return)
                logging.info("Initialize return: {}".format(init_return))

    def _build_server(self, host, port, width, height):

        if self.server is not None:
            return

        if self.server_class.server_type not in self._build.server_types:
            warnings.warn(
                "server_type: %s not available in build: %s, defaulting to WSGI"
                % (self.server_class.server_type, self._build.url)
            )
            self.server_class = ai2thor.wsgi_server.WsgiServer

        if self.server_class == ai2thor.wsgi_server.WsgiServer:
            self.server = ai2thor.wsgi_server.WsgiServer(
                host,
                port=port,
                depth_format=self.depth_format,
                add_depth_noise=self.add_depth_noise,
                width=width,
                height=height,
            )

        elif self.server_class == ai2thor.fifo_server.FifoServer:
            self.server = ai2thor.fifo_server.FifoServer(
                depth_format=self.depth_format,
                add_depth_noise=self.add_depth_noise,
                width=width,
                height=height,
            )

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.stop()

    @property
    def scenes_in_build(self):
        if self._scenes_in_build is not None:
            return self._scenes_in_build

        try:
            event = self.step(action="GetScenesInBuild")
            self._scenes_in_build = set(event.metadata["actionReturn"])
        except ValueError as e:
            # will happen for old builds without GetScenesInBuild
            self._scenes_in_build = set()

        return self._scenes_in_build

    def reset(self, scene=None, **init_params):
        if scene is None:
            scene = self.scene

        if re.match(r"^FloorPlan[0-9]+$", scene):
            scene = scene + "_physics"

        # scenes in build can be an empty set when GetScenesInBuild doesn't exist as an action
        # for old builds
        if self.scenes_in_build and scene not in self.scenes_in_build:

            def key_sort_func(scene_name):
                m = re.search(
                    r"FloorPlan[_]?([a-zA-Z\-]*)([0-9]+)_?([0-9]+)?.*$", scene_name
                )
                last_val = m.group(3) if m.group(3) is not None else -1
                return m.group(1), int(m.group(2)), int(last_val)

            raise ValueError(
                "\nScene '{}' not contained in build (scene names are case sensitive)."
                "\nPlease choose one of the following scene names:\n\n{}".format(
                    scene,
                    ", ".join(sorted(list(self.scenes_in_build), key=key_sort_func)),
                )
            )

        self.server.send(dict(action="Reset", sceneName=scene, sequenceId=0))
        self.last_event = self.server.receive()

        # update the initialization parameters
        init_params = init_params.copy()

        # width and height are updates in 'ChangeResolution', not 'Initialize'
        if ("width" in init_params and init_params["width"] != self.width) or (
            "height" in init_params and init_params["height"] != self.height
        ):
            if "width" in init_params:
                self.width = init_params["width"]
                del init_params["width"]
            if "height" in init_params:
                self.height = init_params["height"]
                del init_params["height"]
            self.step(
                action="ChangeResolution",
                x=self.width,
                y=self.height,
                raise_for_failure=True,
            )

        # updates the initialization parameters
        self.initialization_parameters.update(init_params)

        # RoboTHOR checks
        agent_mode = self.initialization_parameters.get("agentMode", "default")
        if agent_mode.lower() == "bot":
            self.initialization_parameters["agentMode"] = "locobot"
            warnings.warn(
                "On reset and upon initialization, agentMode='bot' has been renamed to agentMode='locobot'."
            )

        if (
            scene in self.robothor_scenes()
            and self.initialization_parameters.get("agentMode", "default").lower()
            != "locobot"
        ):
            warnings.warn(
                "You are using a RoboTHOR scene without using the standard LoCoBot.\n"
                + "Did you mean to mean to set agentMode='locobot' upon initialization or within controller.reset(...)?"
            )

        self.last_event = self.step(
            action="Initialize",
            raise_for_failure=True,
            **self.initialization_parameters,
        )

        self.scene = scene
        return self.last_event

    def random_initialize(
        self,
        random_seed=None,
        randomize_open=False,
        unique_object_types=False,
        exclude_receptacle_object_pairs=[],
        max_num_repeats=1,
        remove_prob=0.5,
    ):

        raise Exception(
            "RandomInitialize has been removed.  Use InitialRandomSpawn - https://ai2thor.allenai.org/ithor/documentation/actions/initialization/#object-position-randomization"
        )

    @lru_cache()
    def ithor_scenes(
        self,
        include_kitchens=True,
        include_living_rooms=True,
        include_bedrooms=True,
        include_bathrooms=True,
    ):
        types = []
        if include_kitchens:
            types.append((1, 31))
        if include_living_rooms:
            types.append((201, 231))
        if include_bedrooms:
            types.append((301, 331))
        if include_bathrooms:
            types.append((401, 431))

        # keep this as a list because the order may look weird otherwise
        scenes = []
        for low, high in types:
            for i in range(low, high):
                scenes.append("FloorPlan%s_physics" % i)
        return scenes

    @lru_cache()
    def robothor_scenes(self, include_train=True, include_val=True):
        # keep this as a list because the order may look weird otherwise
        scenes = []
        stages = dict()

        # from FloorPlan_Train[1:12]_[1:5]
        if include_train:
            stages["Train"] = range(1, 13)
        if include_val:
            # from FloorPlan_Val[1:12]_[1:5]
            stages["Val"] = range(1, 4)

        for stage, wall_configs in stages.items():
            for wall_config_i in wall_configs:
                for object_config_i in range(1, 6):
                    scenes.append(
                        "FloorPlan_{stage}{wall_config}_{object_config}".format(
                            stage=stage,
                            wall_config=wall_config_i,
                            object_config=object_config_i,
                        )
                    )
        return scenes

    @lru_cache()
    def scene_names(self):
        return self.ithor_scenes() + self.robothor_scenes()

    def _prune_release(self, release):
        try:
            # we must try to get a lock here since its possible that a process could still
            # be running with this release
            lock = LockEx(release, blocking=False)
            lock.lock()
            # its possible that another process could prune
            # out a release when running multiple procs
            # that all race to prune the same release
            if os.path.isdir(release):
                tmp_prune_dir = os.path.join(
                    self.tmp_dir,
                    "-".join(
                        [
                            os.path.basename(release),
                            str(time.time()),
                            str(random.random()),
                            "prune",
                        ]
                    ),
                )
                os.rename(release, tmp_prune_dir)
                shutil.rmtree(tmp_prune_dir)

            lock.unlock()
            lock.unlink()
            return True
        except BlockingIOError:
            return False

    def prune_releases(self):
        current_exec_path = self._build.executable_path
        rdir = self.releases_dir
        makedirs(self.tmp_dir)
        makedirs(self.releases_dir)

        # sort my mtime ascending, keeping the 3 most recent, attempt to prune anything older
        all_dirs = list(
            filter(
                os.path.isdir, map(lambda x: os.path.join(rdir, x), os.listdir(rdir))
            )
        )
        dir_stats = defaultdict(lambda: 0)
        for d in all_dirs:
            try:
                dir_stats[d] = os.stat(d).st_mtime
            # its possible for multiple procs to race to stat/prune
            # creating the possibility that between the listdir/stat the directory was
            # pruned
            except FileNotFoundError:
                pass

        sorted_dirs = sorted(all_dirs, key=lambda x: dir_stats[x])[:-3]
        for release in sorted_dirs:
            if current_exec_path.startswith(release):
                continue
            self._prune_release(release)

    def next_interact_command(self):
        # NOTE: Leave this here because it is incompatible with Windows.
        from ai2thor.interact import get_term_character

        current_buffer = ""
        while True:
            commands = self._interact_commands
            current_buffer += get_term_character()
            if current_buffer == "q" or current_buffer == "\x03":
                break

            if current_buffer in commands:
                yield commands[current_buffer]
                current_buffer = ""
            else:
                match = False
                for k, v in commands.items():
                    if k.startswith(current_buffer):
                        match = True
                        break

                if not match:
                    current_buffer = ""

    def interact(
        self,
        semantic_segmentation_frame=False,
        instance_segmentation_frame=False,
        depth_frame=False,
        color_frame=False,
        metadata=False,
    ):
        self.interactive_controller.interact(
            self,
            semantic_segmentation_frame,
            instance_segmentation_frame,
            depth_frame,
            color_frame,
            metadata,
        )

    def multi_step_physics(self, action, timeStep=0.05, max_steps=20):
        events = []
        self.step(action=dict(action="PausePhysicsAutoSim"), raise_for_failure=True)
        events.append(self.step(action))
        while not self.last_event.metadata["isSceneAtRest"]:
            events.append(
                self.step(
                    action=dict(action="AdvancePhysicsStep", timeStep=timeStep),
                    raise_for_failure=True,
                )
            )

            if len(events) == (max_steps - 1):
                events.append(
                    self.step(
                        action=dict(action="UnpausePhysicsAutoSim"),
                        raise_for_failure=True,
                    )
                )
                break

        return events

    def step(self, action=None, **action_args):

        if type(action) is dict:
            action = copy.deepcopy(action)  # prevent changes from leaking
        else:
            action = dict(action=action)

        raise_for_failure = action_args.pop("raise_for_failure", False)
        action.update(action_args)

        if self.headless:
            action["renderImage"] = False

        # prevent changes to the action from leaking
        action = copy.deepcopy(action)

        # XXX should be able to get rid of this with some sort of deprecation warning
        if "AI2THOR_VISIBILITY_DISTANCE" in os.environ:
            action["visibilityDistance"] = float(
                os.environ["AI2THOR_VISIBILITY_DISTANCE"]
            )

        self.last_action = action

        # dangerously converts rotation(float) to rotation(dict(x=0, y=float, z=0))
        # this should be removed when ServerActions have been removed from Unity
        # for all relevant actions.
        rotation = action.get("rotation")
        if rotation is not None and not isinstance(rotation, dict):
            action["rotation"] = dict(y=rotation)

        # Support for deprecated parameter names (old: new)
        # Note that these parameters used to be applicable to ANY action.
        changed_parameter_names = {
            "renderClassImage": "renderSemanticSegmentation",
            "renderObjectImage": "renderInstanceSegmentation",
        }
        for old, new in changed_parameter_names.items():
            if old in action:
                # warnings.warn(old + " has been renamed to " + new)
                action[new] = action[old]
                # not deleting to allow for older builds to continue to work
                # del action[old]

        self.server.send(action)
        try:
            self.last_event = self.server.receive()
        except UnityCrashException as e:
            self.server.stop()
            self.server = None
            # we don't need to pass port or host, since this Exception
            # is only thrown from the FifoServer, start_unity is also
            # not passed since Unity would have to have been started
            # for this to be thrown
            message = "Restarting unity due to crash: %s" % e
            warnings.warn(message)
            self.start(width=self.width, height=self.height, x_display=self.x_display)
            self.reset()
            raise RestartError(message)

        if not self.last_event.metadata["lastActionSuccess"]:
            if self.last_event.metadata["errorCode"] in [
                "InvalidAction",
                "MissingArguments",
                "AmbiguousAction",
                "InvalidArgument",
            ]:
                raise ValueError(self.last_event.metadata["errorMessage"])
            elif raise_for_failure:
                raise RuntimeError(
                    self.last_event.metadata.get("errorMessage", f"{action} failed")
                )

        return self.last_event

    def unity_command(self, width, height, headless):
        command = self._build.executable_path
        if headless:
            command += " -batchmode -nographics"
        else:
            fullscreen = 1 if self.fullscreen else 0
            if QUALITY_SETTINGS[self.quality] == 0:
                raise RuntimeError(
                    "Quality {} is associated with an index of 0. "
                    "Due to a bug in unity, this quality setting would be ignored.".format(
                        self.quality
                    )
                )
            command += (
                " -screen-fullscreen %s -screen-quality %s -screen-width %s -screen-height %s"
                % (fullscreen, QUALITY_SETTINGS[self.quality], width, height)
            )
        return shlex.split(command)

    def _start_unity_thread(self, env, width, height, server_params, image_name):
        # get environment variables

        env["AI2THOR_CLIENT_TOKEN"] = self.server.client_token = str(uuid.uuid4())
        env["AI2THOR_SERVER_TYPE"] = self.server.server_type
        env["AI2THOR_SERVER_SIDE_SCREENSHOT"] = "False" if self.headless else "True"
        for k, v in server_params.items():
            env["AI2THOR_" + k.upper()] = v

        # print("Viewer: http://%s:%s/viewer" % (host, port))
        command = self.unity_command(width, height, headless=self.headless)
        makedirs(self.log_dir)
        self.server.unity_proc = proc = subprocess.Popen(command, env=env, stdout=open(os.path.join(self.log_dir, 'unity.log'), "a"))
        self.unity_pid = proc.pid
        atexit.register(lambda: proc.poll() is None and proc.kill())

    def check_x_display(self, x_display):
        with open(os.devnull, "w") as dn:
            # copying the environment so that we pickup
            # XAUTHORITY values
            env = os.environ.copy()
            env["DISPLAY"] = x_display

            if subprocess.call(["which", "xdpyinfo"], stdout=dn) == 0:
                assert (
                    subprocess.call("xdpyinfo", stdout=dn, env=env, shell=True) == 0
                ), (
                    "Invalid DISPLAY %s - cannot find X server with xdpyinfo"
                    % x_display
                )

    @property
    def tmp_dir(self):
        return os.path.join(self.base_dir, "tmp")

    @property
    def releases_dir(self):
        return os.path.join(self.base_dir, "releases")

    @property
    def cache_dir(self):
        return os.path.join(self.base_dir, "cache")

    @property
    def commits_cache_dir(self):
        return os.path.join(self.cache_dir, "commits")

    @property
    def base_dir(self):
        return os.path.join(os.path.expanduser("~"), ".ai2thor")

    @property
    def log_dir(self):
        return os.path.join(self.base_dir, "log")

    def _cache_commit_filename(self, branch):
        encoded_branch = re.sub(r"[^a-zA-Z0-9_\-.]", "_", re.sub("_", "__", branch))
        return os.path.join(self.commits_cache_dir, encoded_branch + ".json")

    def _cache_commit_history(self, branch, payload):
        makedirs(self.commits_cache_dir)
        cache_filename = self._cache_commit_filename(branch)
        atomic_write(cache_filename, json.dumps(payload))

    def _get_cache_commit_history(self, branch):
        cache_filename = self._cache_commit_filename(branch)
        payload = None
        if os.path.exists(cache_filename):
            with open(cache_filename, "r") as f:
                payload = json.loads(f.read())

        return payload

    def _branch_commits(self, branch):
        import requests

        payload = []
        try:
            res = requests.get(
                "https://api.github.com/repos/allenai/ai2thor/commits?sha=%s" % branch
            )
            if res.status_code == 404:
                raise ValueError("Invalid branch name: %s" % branch)
            elif res.status_code == 200:
                payload = res.json()
                self._cache_commit_history(branch, payload)
            else:
                res.raise_for_status()
        except requests.exceptions.ConnectionError as e:
            payload = self._get_cache_commit_history(branch)
            if payload:
                warnings.warn(
                    "Unable to connect to github.com: %s - using cached commit history for %s"
                    % (e, branch)
                )
            else:
                raise Exception(
                    "Unable to get commit history for branch %s and no cached history exists: %s"
                    % (branch, e)
                )

        return [c["sha"] for c in payload]

    def find_build(self, local_build, commit_id, branch):
        from ai2thor.build import arch_platform_map
        import ai2thor.build

        if platform.architecture()[0] != "64bit":
            raise Exception("Only 64bit currently supported")

        arch = arch_platform_map[platform.system()]

        if branch:
            commits = self._branch_commits(branch)
        elif commit_id:
            ver_build = ai2thor.build.Build(
                arch, commit_id, self.include_private_scenes, self.releases_dir
            )
            ver_build.exists()
            return ver_build
        else:
            git_dir = os.path.normpath(
                os.path.dirname(os.path.realpath(__file__)) + "/../.git"
            )
            commits = (
                subprocess.check_output(
                    "git --git-dir=" + git_dir + " log -n 10 --format=%H", shell=True
                )
                .decode("ascii")
                .strip()
                .split("\n")
            )

        rdir = self.releases_dir
        found_build = None

        if local_build:
            rdir = os.path.normpath(
                os.path.dirname(os.path.realpath(__file__)) + "/../unity/builds"
            )
            commits = [
                "local"
            ] + commits  # we add the commits to the list to allow the ci_build to succeed

        for commit_id in commits:
            commit_build = ai2thor.build.Build(
                arch, commit_id, self.include_private_scenes, rdir
            )

            try:
                if os.path.isdir(commit_build.base_dir) or (
                    not local_build and commit_build.exists()
                ):
                    found_build = commit_build
                    break
            except Exception:
                pass

        if commit_build and commit_build.commit_id != commits[0]:
            warnings.warn(
                "Build for the most recent commit: %s is not available.  Using commit build %s"
                % (commits[0], commit_build.commit_id)
            )

        # print("Got build for %s: " % (found_build.url))
        return found_build

    def start(
        self,
        port=0,
        start_unity=True,
        width=300,
        height=300,
        x_display=None,
        host="127.0.0.1",
        player_screen_width=None,
        player_screen_height=None,
    ):
        self._build_server(host, port, width, height)

        if "AI2THOR_VISIBILITY_DISTANCE" in os.environ:
            import warnings

            warnings.warn(
                "AI2THOR_VISIBILITY_DISTANCE environment variable is deprecated, use \
                the parameter visibilityDistance parameter with the Initialize action instead"
            )

        if player_screen_width is not None:
            warnings.warn(
                "'player_screen_width' parameter is deprecated, use the 'width'"
                " parameter instead."
            )
            width = player_screen_width

        if player_screen_height is not None:
            warnings.warn(
                "'player_screen_height' parameter is deprecated, use the 'height'"
                " parameter instead."
            )
            height = player_screen_height

        if height <= 0 or width <= 0:
            raise Exception("Screen resolution must be > 0x0")

        if self.server.started:
            import warnings

            warnings.warn(
                "start method depreciated. The server started when the Controller was initialized."
            )

            # Stops the current server and creates a new one. This is done so
            # that the arguments passed in will be used on the server.
            self.stop()

        env = os.environ.copy()

        image_name = None

        self.server.start()
        if platform.system() == "Linux":
            if x_display:
                env["DISPLAY"] = ":" + x_display
            elif "DISPLAY" not in env:
                env["DISPLAY"] = ":0.0"

            self.check_x_display(env["DISPLAY"])

        if start_unity:

            self._build.lock_sh()
            self.prune_releases()

            unity_params = self.server.unity_params()
            self._start_unity_thread(env, width, height, unity_params, image_name)

        # receive the first request
        self.last_event = self.server.receive()

        if height < 300 or width < 300:
            self.last_event = self.step("ChangeResolution", x=width, y=height)

        return self.last_event

    def stop(self):
        self.stop_unity()
        self.server.stop()
        self._build.unlock()

    def stop_unity(self):
        if self.unity_pid and process_alive(self.unity_pid):
            self.killing_unity = True
            os.kill(self.unity_pid, signal.SIGTERM)
            for i in range(10):
                if not process_alive(self.unity_pid):
                    break
                time.sleep(0.1)
            if process_alive(self.unity_pid):
                os.kill(self.unity_pid, signal.SIGKILL)


class BFSSearchPoint:
    def __init__(
        self, start_position, move_vector, heading_angle=0.0, horizon_angle=0.0
    ):
        self.start_position = start_position
        self.move_vector = defaultdict(lambda: 0.0)
        self.move_vector.update(move_vector)
        self.heading_angle = heading_angle
        self.horizon_angle = horizon_angle

    def target_point(self):
        x = self.start_position["x"] + self.move_vector["x"]
        z = self.start_position["z"] + self.move_vector["z"]
        return dict(x=x, z=z)


class BFSController(Controller):
    def __init__(self, grid_size=0.25):
        super(BFSController, self).__init__()
        self.rotations = [0, 90, 180, 270]
        self.horizons = [330, 0, 30]
        self.allow_enqueue = True
        self.queue = deque()
        self.seen_points = []
        self.visited_seen_points = []
        self.grid_points = []
        self.grid_size = grid_size
        self._check_visited = False
        self.distance_threshold = self.grid_size / 5.0

    def visualize_points(self, scene_name, wait_key=10):
        import cv2

        points = set()
        xs = []
        zs = []

        # Follow the file as it grows
        for point in self.grid_points:
            xs.append(point["x"])
            zs.append(point["z"])
            points.add(str(point["x"]) + "," + str(point["z"]))

        image_width = 470
        image_height = 530
        image = np.zeros((image_height, image_width, 3), np.uint8)
        if not xs:
            return

        min_x = min(xs) - 1
        max_x = max(xs) + 1
        min_z = min(zs) - 1
        max_z = max(zs) + 1

        for point in list(points):
            x, z = map(float, point.split(","))
            circle_x = round(((x - min_x) / float(max_x - min_x)) * image_width)
            z = (max_z - z) + min_z
            circle_y = round(((z - min_z) / float(max_z - min_z)) * image_height)
            cv2.circle(image, (circle_x, circle_y), 5, (0, 255, 0), -1)

        cv2.imshow(scene_name, image)
        cv2.waitKey(wait_key)

    def has_islands(self):
        queue = []
        seen_points = set()
        mag = self.grid_size

        def enqueue_island_points(p):
            if json.dumps(p) in seen_points:
                return
            queue.append(dict(z=p["z"] + mag, x=p["x"]))
            queue.append(dict(z=p["z"] - mag, x=p["x"]))
            queue.append(dict(z=p["z"], x=p["x"] + mag))
            queue.append(dict(z=p["z"], x=p["x"] - mag))
            seen_points.add(json.dumps(p))

        enqueue_island_points(self.grid_points[0])

        while queue:
            point_to_find = queue.pop()
            for p in self.grid_points:
                dist = math.sqrt(
                    ((point_to_find["x"] - p["x"]) ** 2)
                    + ((point_to_find["z"] - p["z"]) ** 2)
                )

                if dist < 0.05:
                    enqueue_island_points(p)

        return len(seen_points) != len(self.grid_points)

    def build_graph(self):
        import networkx as nx

        graph = nx.Graph()
        for point in self.grid_points:
            self._build_graph_point(graph, point)

        return graph

    def key_for_point(self, point):
        return "{x:0.3f}|{z:0.3f}".format(**point)

    def _build_graph_point(self, graph, point):
        for p in self.grid_points:
            dist = math.sqrt(
                ((point["x"] - p["x"]) ** 2) + ((point["z"] - p["z"]) ** 2)
            )
            if dist <= (self.grid_size + 0.01) and dist > 0:
                graph.add_edge(self.key_for_point(point), self.key_for_point(p))

    def move_relative_points(self, all_points, graph, position, rotation):

        action_orientation = {
            0: dict(x=0, z=1, action="MoveAhead"),
            90: dict(x=1, z=0, action="MoveRight"),
            180: dict(x=0, z=-1, action="MoveBack"),
            270: dict(x=-1, z=0, action="MoveLeft"),
        }

        move_points = dict()

        for n in graph.neighbors(self.key_for_point(position)):
            point = all_points[n]
            x_o = round((point["x"] - position["x"]) / self.grid_size)
            z_o = round((point["z"] - position["z"]) / self.grid_size)
            for target_rotation, offsets in action_orientation.items():
                delta = round(rotation + target_rotation) % 360
                ao = action_orientation[delta]
                action_name = action_orientation[target_rotation]["action"]
                if x_o == ao["x"] and z_o == ao["z"]:
                    move_points[action_name] = point
                    break

        return move_points

    def plan_horizons(self, agent_horizon, target_horizon):
        actions = []
        horizon_step_map = {330: 3, 0: 2, 30: 1, 60: 0}
        look_diff = (
            horizon_step_map[int(agent_horizon)] - horizon_step_map[int(target_horizon)]
        )
        if look_diff > 0:
            for i in range(look_diff):
                actions.append(dict(action="LookDown"))
        else:
            for i in range(abs(look_diff)):
                actions.append(dict(action="LookUp"))

        return actions

    def plan_rotations(self, agent_rotation, target_rotation):
        right_diff = target_rotation - agent_rotation
        if right_diff < 0:
            right_diff += 360
        right_steps = right_diff / 90

        left_diff = agent_rotation - target_rotation
        if left_diff < 0:
            left_diff += 360
        left_steps = left_diff / 90

        actions = []
        if right_steps < left_steps:
            for i in range(int(right_steps)):
                actions.append(dict(action="RotateRight"))
        else:
            for i in range(int(left_steps)):
                actions.append(dict(action="RotateLeft"))

        return actions

    def shortest_plan(self, graph, agent, target):
        import networkx as nx

        path = nx.shortest_path(
            graph,
            self.key_for_point(agent["position"]),
            self.key_for_point(target["position"]),
        )
        actions = []
        all_points = {}

        for point in self.grid_points:
            all_points[self.key_for_point(point)] = point

        # assert all_points[path[0]] == agent['position']

        current_position = agent["position"]
        current_rotation = agent["rotation"]["y"]

        for p in path[1:]:
            inv_pms = {
                self.key_for_point(v): k
                for k, v in self.move_relative_points(
                    all_points, graph, current_position, current_rotation
                ).items()
            }
            actions.append(dict(action=inv_pms[p]))
            current_position = all_points[p]

        actions += self.plan_horizons(agent["cameraHorizon"], target["cameraHorizon"])
        actions += self.plan_rotations(agent["rotation"]["y"], target["rotation"]["y"])
        # self.visualize_points(path)

        return actions

    def enqueue_point(self, point):

        # ensure there are no points near the new point
        if self._check_visited or not any(
            map(
                lambda p: distance(p, point.target_point()) < self.distance_threshold,
                self.seen_points,
            )
        ):
            self.seen_points.append(point.target_point())
            self.queue.append(point)

    def enqueue_points(self, agent_position):

        if not self.allow_enqueue:
            return

        if not self._check_visited or not any(
            map(
                lambda p: distance(p, agent_position) < self.distance_threshold,
                self.visited_seen_points,
            )
        ):
            self.enqueue_point(
                BFSSearchPoint(agent_position, dict(x=-1 * self.grid_size))
            )
            self.enqueue_point(BFSSearchPoint(agent_position, dict(x=self.grid_size)))
            self.enqueue_point(
                BFSSearchPoint(agent_position, dict(z=-1 * self.grid_size))
            )
            self.enqueue_point(
                BFSSearchPoint(agent_position, dict(z=1 * self.grid_size))
            )
            self.visited_seen_points.append(agent_position)

    def search_all_closed(self, scene_name):
        self.allow_enqueue = True
        self.queue = deque()
        self.seen_points = []
        self.visited_seen_points = []
        self.grid_points = []
        event = self.reset(scene_name)
        event = self.step(dict(action="Initialize", gridSize=self.grid_size))
        self.enqueue_points(event.metadata["agent"]["position"])
        while self.queue:
            self.queue_step()
            # self.visualize_points(scene_name)

    def start_search(
        self,
        scene_name,
        random_seed,
        full_grid,
        current_receptacle_object_pairs,
        randomize=True,
    ):

        self.seen_points = []
        self.visited_seen_points = []
        self.queue = deque()
        self.grid_points = []

        # we only search a pre-defined grid with all the cabinets/fridges closed
        # then keep the points that can still be reached
        self.allow_enqueue = True

        for gp in full_grid:
            self.enqueue_points(gp)

        self.allow_enqueue = False

        self.reset(scene_name)
        receptacle_object_pairs = []
        for op in current_receptacle_object_pairs:
            object_id, receptacle_object_id = op.split("||")
            receptacle_object_pairs.append(
                dict(receptacleObjectId=receptacle_object_id, objectId=object_id)
            )

        if randomize:
            self.random_initialize(
                random_seed=random_seed,
                unique_object_types=True,
                exclude_receptacle_object_pairs=receptacle_object_pairs,
            )

        # there is some randomization in initialize scene
        # and if a seed is passed in this will keep it
        # deterministic
        if random_seed is not None:
            random.seed(random_seed)

        self.initialize_scene()
        while self.queue:
            self.queue_step()
            # self.visualize_points(scene_name)

        self.prune_points()
        # self.visualize_points(scene_name)

    # get rid of unreachable points
    def prune_points(self):
        final_grid_points = set()

        for gp in self.grid_points:
            final_grid_points.add(key_for_point(gp["x"], gp["z"]))

        pruned_grid_points = []

        for gp in self.grid_points:
            found = False
            for x in [1, -1]:
                found |= (
                    key_for_point(gp["x"] + (self.grid_size * x), gp["z"])
                    in final_grid_points
                )

            for z in [1, -1]:
                found |= (
                    key_for_point(gp["x"], (self.grid_size * z) + gp["z"])
                    in final_grid_points
                )

            if found:
                pruned_grid_points.append(gp)

        self.grid_points = pruned_grid_points

    def is_object_visible(self, object_id):
        for obj in self.last_event.metadata["objects"]:
            if obj["objectId"] == object_id and obj["visible"]:
                return True
        return False

    def find_visible_receptacles(self):
        receptacle_points = []
        receptacle_pivot_points = []

        # pickup all objects
        visibility_object_id = None
        visibility_object_types = ["Mug", "CellPhone"]
        for obj in self.last_event.metadata["objects"]:
            if obj["pickupable"]:
                self.step(
                    action=dict(
                        action="PickupObject",
                        objectId=obj["objectId"],
                        forceVisible=True,
                    )
                )
            if (
                visibility_object_id is None
                and obj["objectType"] in visibility_object_types
            ):
                visibility_object_id = obj["objectId"]

        for point in self.grid_points:
            self.step(
                dict(action="Teleport", x=point["x"], y=point["y"], z=point["z"]),
                raise_for_failure=True,
            )

            for rot, hor in product(self.rotations, self.horizons):
                event = self.step(
                    dict(action="RotateLook", rotation=rot, horizon=hor),
                    raise_for_failure=True,
                )
                for j in event.metadata["objects"]:
                    if j["receptacle"] and j["visible"]:
                        receptacle_points.append(
                            dict(
                                distance=j["distance"],
                                pivotId=0,
                                receptacleObjectId=j["objectId"],
                                searchNode=dict(
                                    horizon=hor,
                                    rotation=rot,
                                    openReceptacle=False,
                                    pivotId=0,
                                    receptacleObjectId="",
                                    x=point["x"],
                                    y=point["y"],
                                    z=point["z"],
                                ),
                            )
                        )

                        if j["openable"]:
                            self.step(
                                action=dict(
                                    action="OpenObject",
                                    forceVisible=True,
                                    objectId=j["objectId"],
                                ),
                                raise_for_failure=True,
                            )
                        for pivot_id in range(j["receptacleCount"]):
                            self.step(
                                action=dict(
                                    action="Replace",
                                    forceVisible=True,
                                    receptacleObjectId=j["objectId"],
                                    objectId=visibility_object_id,
                                    pivot=pivot_id,
                                ),
                                raise_for_failure=True,
                            )
                            if self.is_object_visible(visibility_object_id):
                                receptacle_pivot_points.append(
                                    dict(
                                        distance=j["distance"],
                                        pivotId=pivot_id,
                                        receptacleObjectId=j["objectId"],
                                        searchNode=dict(
                                            horizon=hor,
                                            rotation=rot,
                                            openReceptacle=j["openable"],
                                            pivotId=pivot_id,
                                            receptacleObjectId=j["objectId"],
                                            x=point["x"],
                                            y=point["y"],
                                            z=point["z"],
                                        ),
                                    )
                                )

                        if j["openable"]:
                            self.step(
                                action=dict(
                                    action="CloseObject",
                                    forceVisible=True,
                                    objectId=j["objectId"],
                                ),
                                raise_for_failure=True,
                            )

        return receptacle_pivot_points, receptacle_points

    def find_visible_objects(self):

        seen_target_objects = defaultdict(list)

        for point in self.grid_points:
            self.step(
                dict(action="Teleport", x=point["x"], y=point["y"], z=point["z"]),
                raise_for_failure=True,
            )

            for rot, hor in product(self.rotations, self.horizons):
                event = self.step(
                    dict(action="RotateLook", rotation=rot, horizon=hor),
                    raise_for_failure=True,
                )

                object_receptacle = dict()
                for obj in event.metadata["objects"]:
                    if obj["receptacle"]:
                        for pso in obj["pivotSimObjs"]:
                            object_receptacle[pso["objectId"]] = obj

                for obj in filter(
                    lambda x: x["visible"] and x["pickupable"],
                    event.metadata["objects"],
                ):

                    # if obj['objectId'] in object_receptacle and\
                    #        object_receptacle[obj['objectId']]['openable'] and not \
                    #        object_receptacle[obj['objectId']]['isopen']:
                    #    continue

                    seen_target_objects[obj["objectId"]].append(
                        dict(distance=obj["distance"], agent=event.metadata["agent"])
                    )

        return seen_target_objects

    def initialize_scene(self):
        self.target_objects = []
        self.object_receptacle = defaultdict(
            lambda: dict(objectId="StartupPosition", pivotSimObjs=[])
        )

        self.open_receptacles = []
        open_pickupable = {}
        pickupable = {}
        is_open = {}

        for obj in filter(
            lambda x: x["receptacle"], self.last_event.metadata["objects"]
        ):
            for oid in obj["receptacleObjectIds"]:
                self.object_receptacle[oid] = obj

            is_open[obj["objectId"]] = obj["openable"] and obj["isOpen"]

        for obj in filter(
            lambda x: x["receptacle"], self.last_event.metadata["objects"]
        ):
            for oid in obj["receptacleObjectIds"]:
                if obj["openable"] or (
                    obj["objectId"] in self.object_receptacle
                    and self.object_receptacle[obj["objectId"]]["openable"]
                ):

                    open_pickupable[oid] = obj["objectId"]
                else:
                    pickupable[oid] = obj["objectId"]

        if open_pickupable.keys():
            self.target_objects = random.sample(open_pickupable.keys(), k=1)
            shuffled_keys = list(open_pickupable.keys())
            random.shuffle(shuffled_keys)
            for oid in shuffled_keys:
                position_target = self.object_receptacle[self.target_objects[0]][
                    "position"
                ]
                position_candidate = self.object_receptacle[oid]["position"]
                dist = math.sqrt(
                    (position_target["x"] - position_candidate["x"]) ** 2
                    + (position_target["y"] - position_candidate["y"]) ** 2
                )
                # try to find something that is far to avoid having the doors collide
                if dist > 1.25:
                    self.target_objects.append(oid)
                    break

        for roid in set(map(lambda x: open_pickupable[x], self.target_objects)):
            if roid in is_open:
                continue
            self.open_receptacles.append(roid)
            self.step(
                dict(action="OpenObject", objectId=roid, forceVisible=True),
                raise_for_failure=True,
            )

    def queue_step(self):
        search_point = self.queue.popleft()
        event = self.step(
            dict(
                action="Teleport",
                x=search_point.start_position["x"],
                y=search_point.start_position["y"],
                z=search_point.start_position["z"],
            )
        )

        if event.metadata["lastActionSuccess"]:
            move_vec = search_point.move_vector
            move_vec["moveMagnitude"] = self.grid_size
            event = self.step(dict(action="Move", **move_vec))
            if event.metadata["agent"]["position"]["y"] > 1.3:
                # pprint(search_point.start_position)
                # pprint(search_point.move_vector)
                # pprint(event.metadata['agent']['position'])
                raise Exception("**** got big point ")

            self.enqueue_points(event.metadata["agent"]["position"])

            if not any(
                map(
                    lambda p: distance(p, event.metadata["agent"]["position"])
                    < self.distance_threshold,
                    self.grid_points,
                )
            ):
                self.grid_points.append(event.metadata["agent"]["position"])

        return event
